#!/usr/bin/env python
#
# Author: Mike McKerns (mmckerns @caltech and @uqfoundation)
# Copyright (c) 2018-2025 The Uncertainty Quantification Foundation.
# License: 3-clause BSD.  The full license text is available at:
#  - https://github.com/uqfoundation/klepto/blob/master/LICENSE
"""
to run this test suite, first build and install `klepto`.

  $ python -m pip install ../..


then run the tests with:

  $ python -m klepto.tests


or, if `nose` is installed:

  $ nosetests  

"""
